package cz.muni.fi.pb162.lambdastreams;

import java.util.*;
import java.util.stream.Collectors;

/**
 * Stream - je sekvencia objektov
 *        - je vhodný na reťazenie viacerých operácii
 *        - nie je dátovou štruktúrou, namiesto toho berie dáta z kolekcii, polí alebo zo vstupu
 *        - neupravuje pôvodné dáta, vracia len ich spracovanú kópiu
 * V streamoch rozpoznávame dva základné druhy operácii - prechodné(transformujú stream) a koncové (vracajú výsledok)
 * Výsledok je "lazy executed", operácie sa vykonajú až keď sa zavolá koncová operácia (napríklad count())
 */
public class Demo {
    private static List<String> numbers = new ArrayList<>(List.of("One", "Two", "Three", "Four", "Five"));
    private static List<String> words = new ArrayList<>(List.of("Dog", "House", "Cat", "Tree", "grill", "goal","pasta", "Tree", "black", "dog"));

    public static void main(String[] args) {
        // v metóde nájdete vysvetlenie lambda funkcie a referencie metód
        // ukážky použitia operácii map, distinct, count
        getNumberOfUniquesIgnoringCase();

        // ukážky použitia operácii skip, limit, filter, sorted, collect
        getFourWordsStartingWithGSortedInList();

        // ukážky streamov s výsledokm typu Optional
        streamWithOptionalAsResult();
    }

    public static void getNumberOfUniquesIgnoringCase() {

        long countUsingLambda = words.stream()

                // lambda je takzvaná anonýmna metóda
                // zapis je ( Type variableName ) -> ( Method body) napríklad (String word) -> word.toLowerCase()
                // ukážková metóda očakáva jeden parameter typu String na ktorom zavolá metódu toLowerCase() a vráti jej výsledok
                // operácia map postupne posúva do lambdy prvky z kolekcie words
                .map((String word) -> word.toLowerCase())

                // operácia distinct() vráti iba unikátne prvky v streame
                .distinct()

                // operácia count() vráti počet prvkov nachádzajúcich sa v streame
                // operácia count() je koncová
                .count();

        long countUsingMethodReference = words.stream()
                // ak chceme nad každým prvkom iba zavolať metódu môžeme labdu nahradiť referenciou metódy
                // referencia metódy má tvar Class::method napríklad String::toLowerCase
                // operácia map tak zavolá nad každým prvkom kolekcie words metódu toLowerCase()
                .map(String::toLowerCase)
                .distinct()
                .count();

        // Ukážky vyššie je možné zapísať následovne bez použitia streamov
        List<String> lowerCaseWords = new ArrayList<>();
        for (String word : words) {
            lowerCaseWords.add(word.toLowerCase());
        }
        Set<String> uniqueWords = new HashSet<>(lowerCaseWords);
        int countViaMethod = uniqueWords.size();

        System.out.println("Vstup: " + words);
        System.out.println("Počet unikátnych slov ignorujúc veľkosť písmen je: " + countUsingLambda
                + ", " + countUsingMethodReference + ", " + countViaMethod);
     }

     public static void getFourWordsStartingWithGSortedInList() {
        List<String> sortedWords = words.stream()
                // operácia skip(number) vynechá zadaný počet prvkov zo začiatku kolekcie
                .skip(2)
                // operácia limit(number) určí maximálny počet prvkov, ktoré sa majú vziať zo streamu
                .limit(4)
                // operácia filter zahodí prvky ktoré nesplnia podmienku
                // podmienka je daná lambda funkciou alebo referenciou metódy, ktorá vracia boolean
                .filter((String word) -> word.startsWith("g"))
                // operácia sorted() zoradí prvky v kolekcii
                .sorted()
                // operácia collect() zozbiera zvyšné prvky a podľa parametra ich vloží do kolekcie alebo do poľa
                // Collectors je statická trieda, ktorá ponúka redukčné operácie, napríklad akumulovanie prvkov do kolekcii
                .collect(Collectors.toList());

        System.out.println("Vstup: " + words);
        System.out.println("Maximálne štyri slová začínajúce na g, zoradené podľa abecedy a vložené do listu: " + sortedWords );
     }

     public static void streamWithOptionalAsResult() {
         // OptionalDouble je kontajnerový objekt, ktorý môže ale nemusí obsahovať hodnotu double
         // typ Optional je najmä využívaný ako návratový typ v metóde, ktorá nemusí mať výsledok (napríklad delenie 0)
        OptionalDouble optional = numbers.stream()
                // operácia mapToInt je veľmi podobná operácii map, s rozdielom, že zaručuje že výstupný stream bude typu IntStream
                .mapToInt(String::length)
                // operácia average() nám zráta priemer z číselného streamu
                .average();

         // metóda orElse(variable) vráti hodnotu objektu optional
         // ak objekt nemá hodnotu, vráti parameter variable metódy orElse() (v tomto prípade by to bola 0)
        System.out.println("Priemerná dĺžka slov v zozname " + numbers + " je " + optional.orElse(0));

         // výsledok bude určite bez hodnoty lebo podmienku operácie filter nesplní žiadne slovo
         // tým by pri výpočte priemeru dochádzalo k delení 0 -> výsledok priemeru nemá hodnotu
        OptionalDouble optionalWithoutValue = numbers.stream()
                .filter((String number) -> number.length() < -10)
                .mapToInt(String::length)
                .average();

        // metóda isPresent() vráti true ak má objekt typu Optional hodnotu, false ak nemá
        System.out.println("Má objekt optional hodnotu: " + optionalWithoutValue.isPresent());
        // vráti sa hodnota ktorú sme predali ako parameter metódy orElse()
        System.out.println("Priemerná dlžka slov kratších ako -10 znakov je: " + optionalWithoutValue.orElse(0));
     }
}
